"""The module contains a set of API for HTTP sessions."""
from abc import abstractmethod
from types import TracebackType
from typing import Any, AsyncContextManager, Optional, Type
import requests
from punish import AbstractStyle
from requests.auth import HTTPBasicAuth
from aiorequest.types import OptionalAnyDict, OptionalStr
from aiorequest.responses import HttpResponse, Response, safe_response
from aiorequest.urls import Address


class Session(AbstractStyle, AsyncContextManager["Session"]):
    """The class represents abstract interfaces for an API Session."""

    @abstractmethod
    async def get(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``GET`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def options(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``OPTIONS`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def head(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``HEAD`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def post(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """Performs ``POST`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def put(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """Performs ``PUT`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def patch(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """Performs ``PATCH`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments
        Returns: response element
        """
        pass

    @abstractmethod
    async def delete(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``DELETE`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments
        Returns: response element
        """
        pass


class HttpSession(Session):
    """The class provides interfaces for current API HTTP session."""

    def __init__(self, session: requests.Session = requests.Session()) -> None:
        self._session: requests.Session = session

    async def __aenter__(self) -> Session:
        """See base class."""
        return self

    async def get(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await safe_response(HttpResponse(self._session.get(await url.as_str(), **kwargs)))

    async def options(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await safe_response(
            HttpResponse(self._session.options(await url.as_str(), **kwargs))
        )

    async def head(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await safe_response(HttpResponse(self._session.head(await url.as_str(), **kwargs)))

    async def post(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await safe_response(
            HttpResponse(self._session.post(str(url), data=plain, json=as_dict, **kwargs))
        )

    async def put(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await safe_response(
            HttpResponse(self._session.put(str(url), data=plain, json=as_dict, **kwargs))
        )

    async def patch(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await safe_response(
            HttpResponse(self._session.patch(str(url), data=plain, json=as_dict, **kwargs))
        )

    async def delete(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await safe_response(HttpResponse(self._session.delete(str(url), **kwargs)))

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """See base class."""
        self._session.close()


class LoggedHttpSession(Session):
    """The class provides logged HTTP session."""

    def __init__(
        self, username: str, password: str, session: requests.Session = requests.Session()
    ) -> None:
        session.auth = HTTPBasicAuth(username, password)
        self._session: Session = HttpSession(session)

    async def __aenter__(self) -> Any:
        """See base class."""
        return await self._session.__aenter__()

    async def get(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await self._session.get(url, **kwargs)

    async def options(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await self._session.options(url, **kwargs)

    async def head(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await self._session.head(url, **kwargs)

    async def post(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await self._session.post(url, plain, as_dict, **kwargs)

    async def put(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await self._session.post(url, plain, as_dict, **kwargs)

    async def patch(
        self,
        url: Address,
        plain: OptionalStr = None,
        as_dict: OptionalAnyDict = None,
        **kwargs: Any,
    ) -> Response:
        """See base class."""
        return await self._session.patch(url, plain, as_dict, **kwargs)

    async def delete(self, url: Address, **kwargs: Any) -> Response:
        """See base class."""
        return await self._session.delete(url, **kwargs)

    async def __aexit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """See base class."""
        await self._session.__aexit__(exception_type, exception_value, traceback)



"""The module provides API for credentials."""
from abc import abstractmethod
from typing import Any, Dict, Optional, Union
from punish import AbstractStyle

AnyDict = Dict[Any, Any]
OptionalStr = Optional[str]
OptionalAnyDict = Optional[AnyDict]
AnyUnionDict = Union[AnyDict, Any]


class Credentials(AbstractStyle):
    """The class represents abstraction for credentials."""

    @property
    @abstractmethod
    async def username(self) -> str:
        """Returns a username."""
        pass

    @property
    @abstractmethod
    async def password(self) -> str:
        """Returns a password."""
        pass

    @abstractmethod
    async def as_str(self) -> str:
        """Returns credentials string representation."""
        pass


class AuthCredentials(Credentials):
    """The class represents authorization credentials."""

    def __init__(self, username: str, password: str) -> None:
        self._username: str = username
        self._password: str = password

    @property
    async def username(self) -> str:
        """Returns auth username."""
        return self._username

    @property
    async def password(self) -> str:
        """Returns auth password."""
        return self._password

    async def as_str(self) -> str:
        """Returns auth credentials string representation."""
        return (
            f"{self.__class__.__name__}"
            f"(username='{await self.username}', password='{await self.password}')"
        )



from http import HTTPStatus
from aiorequest.responses import JsonType, Response


class FakeHttpResponse(Response):
    """The class represents fake HTTP response interface."""

    def __init__(
        self, code: HTTPStatus, is_ok: bool = True, as_str: str = str(), as_dict: JsonType = {}
    ) -> None:
        self._code: HTTPStatus = code
        self._is_ok: bool = is_ok
        self._as_str: str = as_str
        self._as_dict: JsonType = as_dict

    async def is_ok(self) -> bool:
        return self._is_ok

    async def code(self) -> HTTPStatus:
        return self._code

    async def as_json(self) -> JsonType:
        return self._as_dict

    async def as_str(self) -> str:
        return self._as_str



from http import HTTPStatus
from typing import Iterable
import pytest
from tests.fake import FakeHttpResponse
from aiorequest.responses import Response, ResponseError, safe_response
from tests.markers import asyncio, unit

pytestmark = [unit, asyncio]


@pytest.mark.parametrize(  # noqa: PT006, PT007
    "code, expected",
    (
        pytest.param(
            HTTPStatus.CONTINUE,
            (HTTPStatus.CONTINUE, HTTPStatus.SWITCHING_PROTOCOLS, HTTPStatus.PROCESSING),
            id="info",
        ),
        pytest.param(
            HTTPStatus.OK,
            (HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.UNAVAILABLE_FOR_LEGAL_REASONS),
            id="success",
        ),
        pytest.param(
            HTTPStatus.MULTIPLE_CHOICES,
            (HTTPStatus.MULTIPLE_CHOICES, HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND),
            id="redirect",
        ),
        pytest.param(
            HTTPStatus.BAD_REQUEST,
            (HTTPStatus.BAD_REQUEST, HTTPStatus.UNAUTHORIZED, HTTPStatus.PAYMENT_REQUIRED),
            id="client error",
        ),
        pytest.param(
            HTTPStatus.INTERNAL_SERVER_ERROR,
            (HTTPStatus.INTERNAL_SERVER_ERROR, HTTPStatus.NOT_IMPLEMENTED, HTTPStatus.BAD_GATEWAY),
            id="server error",
        ),
    ),
)
async def test_safe_response_code(code: HTTPStatus, expected: Iterable[int]) -> None:
    assert isinstance(await safe_response(FakeHttpResponse(code), success_codes=expected), Response)


async def test_safe_response_error() -> None:
    with pytest.raises(ResponseError):
        await safe_response(
            FakeHttpResponse(HTTPStatus.INTERNAL_SERVER_ERROR),
            success_codes=(HTTPStatus.OK, HTTPStatus.CREATED),
        )


async def test_response_as_json(response: Response) -> None:
    assert await response.as_json()


async def test_response_is_ok(response: Response) -> None:
    assert await response.is_ok()


async def test_response_code(response: Response) -> None:
    assert await response.code() is HTTPStatus.OK


async def test_response_text(response: Response) -> None:
    assert await response.as_str()


async def test_logged_response_as_json(logged_response: Response) -> None:
    assert await logged_response.as_json()


async def test_logged_response_is_ok(logged_response: Response) -> None:
    assert await logged_response.is_ok()


async def test_logged_response_code(logged_response: Response) -> None:
    assert await logged_response.code() is HTTPStatus.OK


async def test_logged_response_text(logged_response: Response) -> None:
    assert await logged_response.as_str()



import pytest
from aiorequest.types import AuthCredentials, Credentials
from aiorequest.responses import Response
from aiorequest.sessions import HttpSession, LoggedHttpSession, Session
from aiorequest.urls import Address, HttpUrl


@pytest.fixture(scope="session")
async def session_url() -> Address:
    yield HttpUrl(host="xkcd.com", path="info.0.json")


@pytest.fixture(scope="session")
async def credentials() -> Credentials:
    yield AuthCredentials(username="superuser", password="superpass")


@pytest.fixture(scope="session")
async def session() -> Session:
    http_session: Session
    async with HttpSession() as http_session:
        yield http_session


@pytest.fixture(scope="session")
async def logged_session(credentials: Credentials) -> Session:
    logged_http_session: Session
    async with LoggedHttpSession(
        await credentials.username, await credentials.password
    ) as logged_http_session:
        yield logged_http_session


@pytest.fixture(scope="session")
async def response(session: Session, session_url: Address) -> Response:
    yield await session.get(session_url)


@pytest.fixture(scope="session")
async def logged_response(logged_session: Session, session_url: Address) -> Response:
    yield await logged_session.get(session_url)



from aiorequest import Credentials
from tests.markers import asyncio, unit

pytestmark = [unit, asyncio]


async def test_username(credentials: Credentials) -> None:
    assert await credentials.username == "superuser"


async def test_password(credentials: Credentials) -> None:
    assert await credentials.password == "superpass"


async def test_as_string(credentials: Credentials) -> None:
    assert (
        await credentials.as_str() == "AuthCredentials(username='superuser', password='superpass')"
    )



"""The module contains a set of API for HTTP responses types."""
from typing import Iterable
from http import HTTPStatus
import requests
from punish import AbstractStyle, abstractstyle
from aiorequest.types import AnyUnionDict

JsonType = AnyUnionDict


class ResponseError(Exception):
    """The class represents HTTP api request response error."""

    pass


class Response(AbstractStyle):
    """The class represents an abstraction of a response from an API request."""

    @abstractstyle
    async def is_ok(self) -> bool:
        """Returns `True` if response is `OK` otherwise `False`."""
        pass

    @abstractstyle
    async def code(self) -> HTTPStatus:
        """Returns HTTP response status code."""
        pass

    @abstractstyle
    async def as_json(self) -> JsonType:
        """Returns HTTP response data as dictionary type."""
        pass

    @abstractstyle
    async def as_str(self) -> str:
        """Returns HTTP response data as plain data type."""
        pass


class HttpResponse(Response):
    """The class represents an HTTP response from HTTP API request."""

    def __init__(self, response: requests.Response) -> None:
        self._response: requests.Response = response

    async def is_ok(self) -> bool:
        """See base class."""
        return self._response.ok

    async def code(self) -> HTTPStatus:
        """See base class."""
        return HTTPStatus(self._response.status_code)

    async def as_json(self) -> JsonType:
        """See base class."""
        return self._response.json()

    async def as_str(self) -> str:
        """See base class."""
        return self._response.text


async def safe_response(
    response: Response,
    success_codes: Iterable[int] = (HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.NO_CONTENT),
) -> Response:
    """Specifies safe response from iterable of success HTTP status codes.

    Args:
        response: a specific HTTP response
        success_codes: a list of allowed success codes

    Raises:
        `ResponseError` if HTTP response contains a set of errors
    Returns: a response
    """
    if await response.code() not in success_codes:
        raise ResponseError(
            f"HTTP response contains some errors with '{await response.code()}' status code! "
            f"Reason: {await response.as_str()}"
        )
    return response



import pytest
from aiorequest.urls import Address, HttpUrl, HttpsUrl, Url
from tests.markers import asyncio, unit

_host: str = "9.9.9.9"
_path: str = "/api/path"

pytestmark = [unit, asyncio]


@pytest.fixture(scope="module")
async def url() -> Address:
    yield Url(_host, protocol="ftp", path=_path)


@pytest.fixture(scope="module")
async def http_url() -> Address:
    yield HttpUrl(_host, _path)


@pytest.fixture(scope="module")
async def https_url() -> Address:
    yield HttpsUrl(_host, _path)


async def test_url_with_leading_slash() -> None:
    assert await Url(_host, protocol="ftp", path="/w/r").as_str() == f"ftp://{_host}/w/r"


async def test_url_without_leading_slash() -> None:
    assert await Url(_host, protocol="ftp", path="w/r").as_str() == f"ftp://{_host}/w/r"


async def test_url_with_full_host_path() -> None:
    assert await Url(f"ftp://{_host}/w/r", protocol="ftp").as_str() == f"ftp://{_host}/w/r"


async def test_url_host(url: Address) -> None:
    assert await url.host() == _host


async def test_url_matcher(url: Address) -> None:
    assert await url.matcher() == _path


async def test_url_as_str(url: Address) -> None:
    assert await url.as_str() == f"ftp://{_host}{_path}"


async def test_http_url_host(http_url: Address) -> None:
    assert await http_url.host() == _host


async def test_http_url_matcher(http_url: Address) -> None:
    assert await http_url.matcher() == _path


async def test_http_url_as_str(http_url: Address) -> None:
    assert await http_url.as_str() == f"http://{_host}{_path}"


async def test_https_url_host(https_url: Address) -> None:
    assert await https_url.host() == _host


async def test_https_url_matcher(https_url: Address) -> None:
    assert await https_url.matcher() == _path


async def test_https_url_as_str(https_url: Address) -> None:
    assert await https_url.as_str() == f"https://{_host}{_path}"
